from typing import Tuple

from .Spawner import Spawner
from ..Stage import Stage


class DockerSpawner(Spawner):
    def __init__(self, image: str, server_name: str = None):
        super().__init__(server_name)
        self.image: str = image

    def url(self, successors: Tuple[Stage, ...]) -> str:
        return self._get(successors, image=self.image)



import asyncio
from json import dumps as json_encode
from typing import IO

from .ResourceFile import ResourceFile
from .. import Moodle
from ..error import MoodleParseError


class ResourceList:
    def __init__(self, md: Moodle,
                 session_key: str, item_id: str, client_id: str, context_id: str,
                 file_path: str, full_name: str,
                 files: list[ResourceFile], children: list['ResourceList']):
        self._md: Moodle = md

        self._session_key: str = session_key
        self._item_id: str = item_id
        self._client_id: str = client_id
        self._context_id: str = context_id

        self.file_path: str = file_path
        self.full_name: str = full_name
        self.files: list[ResourceFile] = files
        self.children: list['ResourceList'] = children

    @staticmethod
    async def fetch(md: Moodle,
                    session_key: str, item_id: str, client_id: str, context_id: str, file_path: str) -> 'ResourceList':
        json = await md.session.post_json(
            f'/repository/draftfiles_ajax.php?action=list',
            {
                'sesskey': session_key,
                'itemid': item_id,
                'client_id': client_id,
                'filepath': file_path
            }
        )

        return ResourceList(
            md,
            session_key, item_id, client_id, context_id,
            json['path'][-1]['path'], json['path'][-1]['name'],
            [
                ResourceFile(
                    md,
                    session_key, item_id, client_id,
                    file_path, file['filename'], file['fullname'], file['size'], file['mimetype'],
                    file['author'], file['license'], file['datecreated'], file['datemodified'], file['isref'],
                    file['url']
                )
                for file in json['list']
                if file['type'] == 'file'
            ],
            await asyncio.gather(*(
                ResourceList.fetch(
                    md, session_key, item_id, client_id, context_id, file['filepath']
                )
                for file in json['list']
                if file['type'] == 'folder'
            ))
        )

    async def delete(self):
        json = await self._md.session.post_json(
            '/repository/draftfiles_ajax.php?action=deleteselected',
            {
                'sesskey': self._session_key,
                'client_id': self._client_id,
                'filepath': self.file_path,
                'itemid': self._item_id,
                'selected': json_encode([{
                    'filepath': self.file_path,
                    'filename': '.'
                }])
            }
        )

        if 'error' in json:
            raise MoodleParseError(json['error'])

    async def upload(self, file: IO, name: str, author: str, overwrite: bool = True) -> bool:
        json = await self._md.session.post_json(
            '/repository/repository_ajax.php?action=upload',
            {
                'sesskey': self._session_key,
                'repo_id': str(5),  # always 5?
                'itemid': self._item_id,
                'author': author,
                'savepath': self.file_path,
                'title': name,
                'overwrite': str(1 if overwrite else 0),
                'ctx_id': self._context_id,
                'repo_upload_file': file
            },
            no_check_mimetype=True
        )

        if 'error' in json:
            raise MoodleParseError(json['error'])

        return 'event' not in json or json['event'] != 'fileexists'



from .Activity import Activity


class LTIActivity(Activity):
    async def set_tool_url(self, tool_url: str, save: bool = False):
        return await self._set(save, toolurl=tool_url)

    # TODO typeid



from json import dumps as json_encode

from .. import Moodle
from ..error import MoodleParseError


class ResourceFile:
    def __init__(self, md: Moodle,
                 session_key: str, item_id: str, client_id: str,
                 file_path: str, file_name: str, full_name: str, size: int, mime_type: str,
                 author: str, license: str, created: int, modified: int, is_ref: bool,
                 url: str):
        self._md: Moodle = md

        self._session_key: str = session_key
        self._item_id: str = item_id
        self._client_id: str = client_id
        self._file_path: str = file_path

        self.file_name: str = file_name
        self.full_name: str = full_name
        self.size: int = size
        self.mime_type: str = mime_type
        self.author: str = author
        self.license: str = license
        self.created: int = created
        self.modified: int = modified
        self.is_ref: bool = is_ref
        self.url: str = url

    async def delete(self):
        json = await self._md.session.post_json(
            '/repository/draftfiles_ajax.php?action=deleteselected',
            {
                'sesskey': self._session_key,
                'client_id': self._client_id,
                'filepath': self._file_path,
                'itemid': self._item_id,
                'selected': json_encode([{
                    'filepath': self._file_path,
                    'filename': self.file_name
                }])
            }
        )

        if 'error' in json:
            raise MoodleParseError(json['error'])



from typing import Tuple

from ..Stage import Stage


class JupyterHub(Stage):
    def __init__(self, name: str):
        super().__init__()
        self.name: str = name

    def url(self, successors: Tuple[Stage, ...]) -> str:
        url = f'https://{self.name}/'

        if len(successors) > 0:
            url += f'?next={successors[0].encoded_url(successors[1:])}'

        return url



from typing import Tuple

from ..Stage import Stage


class NBGitPuller(Stage):
    def __init__(self, repo: str, branch: str, target_path: str):
        super().__init__()
        self.repo: str = repo
        self.branch: str = branch
        self.target_path: str = target_path

    def url(self, successors: Tuple[Stage, ...]) -> str:
        repo = self.encode(self.repo)
        branch = self.encode(self.branch)
        target_path = self.encode(self.target_path)

        path = f'git-pull?repo={repo}&branch={branch}&targetPath={target_path}'

        if len(successors) > 0:
            path += f'&urlpath={successors[0].encoded_url(successors[1:])}'

        return path



import asyncio
import json
import logging
import os
from pathlib import Path
from sys import argv

import yaml

from .util.lti import generate as generate_lti_url
from . import Moodle, ResourceActivity, LTIActivity


async def main():
    # check environment variables
    if 'MOODLE_URL' not in os.environ:
        raise ValueError('MOODLE_URL environment variable not set')
    if 'MOODLE_USERNAME' not in os.environ:
        raise ValueError('MOODLE_USERNAME environment variable not set')
    if 'MOODLE_PASSWORD' not in os.environ:
        raise ValueError('MOODLE_PASSWORD environment variable not set')

    MOODLE_URL = os.environ['MOODLE_URL']
    MOODLE_USERNAME = os.environ['MOODLE_USERNAME']
    MOODLE_PASSWORD = os.environ['MOODLE_PASSWORD']
    MOODLE_IDP = os.environ.get('MOODLE_IDP', None)
    MOODLE_REQUEST_WAIT_TIME = float(os.environ.get('MOODLE_REQUEST_WAIT_TIME', '1.0'))

    # read mapping
    if len(argv) != 2:
        raise ValueError('Usage: python -m moodle_api path/to/mapping.[yaml|json]')
    mapping_path = argv[1]

    if mapping_path.endswith(('.yml', '.yaml')):
        logging.warning(f'read yaml mapping from "{mapping_path}"')
        with open(mapping_path, 'r', encoding='utf-8') as mapping_file:
            mapping_source = yaml.load(mapping_file, Loader=yaml.Loader)
    elif mapping_path.endswith('.json'):
        logging.warning(f'read json mapping from "{mapping_path}"')
        with open(mapping_path, 'r', encoding='utf-8') as mapping_file:
            mapping_source = json.load(mapping_file)

    # convert mapping
    mapping: dict[int, dict] = {}

    for activity, options in mapping_source.items():
        if options is None or len(options) == 0:
            logging.warning(f'skip empty activity "{activity}"')
            continue

        if isinstance(activity, int):
            key = activity
        elif isinstance(activity, str) and activity.isnumeric():
            key = int(activity)
        elif isinstance(activity, str) and activity.startswith(MOODLE_URL):
            key = int(activity.rsplit('id=', 1)[-1])
        else:
            raise ValueError(f'unknown activity "{activity}"')

        if not isinstance(options, dict):
            raise ValueError(f'unknown options type "{type(options)}" for activity {activity}')

        value = {}
        for k, v in options.items():
            match k:
                case 'name':
                    if not isinstance(v, str):
                        raise ValueError(f'invalid type "{type(v)}" for {k} in activity {activity}')

                    value[k] = v

                case 'description':
                    if not isinstance(v, str):
                        raise ValueError(f'invalid type "{type(v)}" for {k} in activity {activity}')

                    value[k] = v

                case 'file':
                    if isinstance(v, str):
                        path = Path(v)
                        value[k] = {
                            'clear': False,
                            'path': path,
                            'name': v.split('/')[-1],
                            'author': ''
                        }
                    elif isinstance(v, dict):
                        if 'path' not in v:
                            raise ValueError(f'missing path for file update in activity {activity}')

                        for l in v:
                            if l not in ('clear', 'path', 'name', 'author'):
                                raise ValueError(f'unknown option "{l}" for file update in activity {activity}')

                        path = Path(v['path'])
                        value[k] = {
                            'clear': v.get('clear', False),
                            'path': path,
                            'name': v.get('name', v['path'].split('/')[-1]),
                            'author': v.get('author', '')
                        }
                    else:
                        raise ValueError(f'invalid type "{type(v)}" for {k} in activity {activity}')

                    if not path.exists():
                        raise ValueError(f'invalid path "{path}" for file update in {activity}')

                case 'lti':
                    value[k] = generate_lti_url(
                        v['hub'],
                        v['image'],
                        v['git']['repository'], v['git']['branch'], v['git']['target'],
                        v['start']
                    )

                case _:
                    raise ValueError(f'unknown option "{key}" for activity {activity}')

        mapping[key] = value

    # create moodle
    async with Moodle(MOODLE_URL, MOODLE_REQUEST_WAIT_TIME) as moodle:
        # login
        if MOODLE_IDP is None:
            logging.warning(f'login to "{MOODLE_URL}" as "{MOODLE_USERNAME}"')
            await moodle.login(MOODLE_USERNAME, MOODLE_PASSWORD)
        else:
            logging.warning(f'login to "{MOODLE_URL}" as "{MOODLE_USERNAME}" using IDP "{MOODLE_IDP}"')
            await moodle.shibboleth_login(MOODLE_USERNAME, MOODLE_PASSWORD, MOODLE_IDP)

        # update activities
        for activity_id, options in mapping.items():
            logging.warning(f'processing activity {activity_id}')
            activity = await moodle.activity(activity_id).fetch()

            for option, value in options.items():
                match option:
                    case 'name':
                        logging.info(f'set name="{value}" for activity {activity_id}')
                        await activity.set_name(value)

                    case 'description':
                        logging.info(f'set description="{value}" for activity {activity_id}')
                        await activity.set_description(value)

                    case 'file':
                        if not isinstance(activity, ResourceActivity):
                            raise ValueError(f'invalid type "{type(activity)}" for activity {activity_id}')

                        logging.info(f'receive files for activity {activity_id}')
                        resource_list = await activity.resource_list()

                        if value['clear']:
                            logging.info(f'clear files for activity {activity_id}')

                            for folder in resource_list.children:
                                await folder.delete()
                            for file in resource_list.files:
                                await file.delete()

                        with open(value['path'], 'rb') as file:
                            logging.info(f'upload "{value["path"]}" for activity {activity_id}')
                            await resource_list.upload(file, value['name'], value['author'])

                    case 'lti':
                        if not isinstance(activity, LTIActivity):
                            raise ValueError(f'invalid type "{type(activity)}" for activity {activity_id}')

                        logging.info(f'set lti="{value}" for activity {activity_id}')
                        await activity.set_tool_url(value)

            await activity.set()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    asyncio.run(main())



from time import time
from typing import List

from . import Assignment
from .SubmissionFile import SubmissionFile
from .. import Moodle


class Submission:
    def __init__(self, md: Moodle, assignment: Assignment, user_id: int, files: List[SubmissionFile],
                 grade_attempt: int, session_key: str):
        self._md: Moodle = md
        self._assignment: Assignment = assignment
        self._user_id: int = user_id
        self._files: List[SubmissionFile] = files
        self._grade_attempt: int = grade_attempt
        self._session_key: str = session_key

    @property
    def assignment(self) -> Assignment:
        return self._assignment

    @property
    def user_id(self) -> int:
        return self._user_id

    @property
    def files(self) -> List[SubmissionFile]:
        return self._files

    async def lock(self):
        await self._md.session.get_bytes('/mod/assign/view.php'
                                         f'?id={self.assignment.id}'
                                         f'&userid={self.user_id}'
                                         f'&sesskey={self._session_key}'
                                         f'&action=lock')

    async def grade(self, score: float, comment: str, send_notification: bool = True):
        await self._md.session.post_html('/mod/assign/view.php', {
            'id': self._assignment.id,
            'sesskey': self._session_key,
            'action': 'quickgrade',
            'lastpage': '',
            'content': 'Kommentar hinzufügen...',
            'savequickgrades': 'Bewertungsänderungen sichern',
            'sendstudentnotifications': 1 if send_notification else 0,
            '_qf__mod_assign_quick_grading_form': 1,
            f'grademodified_{self.user_id}': int(time()),
            f'gradeattempt_{self.user_id}': self._grade_attempt,
            f'quickgrade_{self.user_id}': score,
            f'quickgrade_comments_{self.user_id}': comment
        })



import re

from .activity import Activity
from .error import MoodleLoginError
from .assignment import Assignment
from .util import LimitedSession


class Moodle:
    def __init__(self, host: str, wait_time: float = 1.0):
        self.session: LimitedSession = LimitedSession(host, wait_time)

    # context manager
    async def __aenter__(self):
        await self.session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.session.__aexit__(exc_type, exc_val, exc_tb)

    # login
    async def logged_in(self) -> bool:
        page = await self.session.get_html('/login/index.php')
        text = page.find(id='modal-body')

        if text is not None:
            if re.match(r'^Sie sind bereits als (.*?) angemeldet. '
                        r'Sie müssen sich abmelden, bevor Sie sich mit anderen Daten wieder anmelden.$',
                        text.text):
                return True

        return False

    async def login(self, username: str, password: str):
        # receive login token
        page = await self.session.get_html('/login/index.php')

        input = page.select_one('input[name="logintoken"]')
        login_token = input['value']

        # send login request
        page = await self.session.post_html('/login/index.php', {
            'logintoken': login_token,
            'username': username,
            'password': password
        })

        error_message = page.find(id='loginerrormessage')
        if error_message is not None:
            raise MoodleLoginError(error_message.text)

    async def shibboleth_login(self, username: str, password: str, provider_id: str):
        # receive idp login page
        page = await self.session.post_html('/Shibboleth.sso/Login', {
            'target': f'{self.session.host}/auth/shibboleth/index.php',
            'providerId': f'{provider_id}/idp/shibboleth'
        })

        # send login
        form = page.select_one('form[method="post"]')
        action = form['action']

        input = form.select_one('input[name="csrf_token"]')
        csrf_token = input['value']

        page = await self.session.post_html(f'{provider_id}{action}', {
            'csrf_token': csrf_token,
            'j_username': username,
            'j_password': password,
            '_eventId_proceed': ''
        })

        error_message = page.select_one('p.output--error')
        if error_message is not None:
            raise MoodleLoginError(error_message.text)

        # handle redirect form
        form = page.select_one('form[method="post"]')
        action = form['action']

        input = form.select_one('input[name="RelayState"]')
        relay_state = input['value']

        input = form.select_one('input[name="SAMLResponse"]')
        saml_response = input['value']

        await self.session.post_html(action, {
            'RelayState': relay_state,
            'SAMLResponse': saml_response
        })

    async def is_logged_in(self) -> bool:
        page, redirects = await self.session.get_html_redirects('/my')

        if not redirects[-1][1].path.endswith(('/my', '/my/')):
            return False

        if 'Sie sind als Gast angemeldet' in page.text:
            return False

        return True

    # assignments
    def assigment(self, id: int) -> Assignment:
        return Assignment(self, id)

    # TODO course, chapters?

    # activity
    def activity(self, id: int) -> Activity:
        return Activity(self, id)



import urllib.parse
from typing import Optional, Tuple


class Stage:
    def __init__(self):
        self.parent: Optional[Stage] = None

    @staticmethod
    def encode(value: str) -> str:
        return urllib.parse.quote(value, safe='')

    def encoded_url(self, successors: Tuple['Stage', ...]) -> str:
        return self.encode(self.url(successors))

    def url(self, successors: Tuple['Stage', ...]) -> str:
        raise NotImplementedError

    def __str__(self, successors: Tuple['Stage', ...] = ()) -> str:
        if self.parent is not None:
            return self.parent.__str__((self,) + successors)
        else:
            return self.url(successors)

    def __add__(self, other: 'Stage') -> 'Stage':
        other.parent = self
        return other



from .Activity import Activity
from .ResourceList import ResourceList
from ..error import MoodleParseError


class ResourceActivity(Activity):
    async def resource_list(self) -> ResourceList:
        # session key and files id can be obtained from the form
        session_key = self.session_key
        files = self['files']

        # for the client id, we need to find the file manager div
        for div in self._page.select('div.filemanager'):
            if div.attrs['id'].startswith('filemanager-'):
                client_id = div.attrs['id'][len('filemanager-'):]
                break
        else:
            raise MoodleParseError(f'could not extract client id from file manager')

        # for the context id, we need to find an integrated object
        context_id = None
        for object in self._page.select('object[type="text/html"]'):
            for attr in object.attrs['data'].split('&'):
                if attr.startswith(('ctx_id', 'amp;ctx_id')):
                    _, context_id = attr.split('=', 1)
                    break

        if context_id is None:
            raise MoodleParseError(f'could not extract client id from file manager')

        # fetch tree
        return await ResourceList.fetch(self._md, session_key, files, client_id, context_id, '/')



from os import PathLike
from typing import Union

"""
import aiofiles
from aiofiles.threadpool.binary import AsyncBufferedIOBase
"""

from .. import Moodle


class SubmissionFile:
    def __init__(self, md: Moodle, name: str, url: str):
        self._md: Moodle = md
        self._name: str = name
        self._url: str = url

    @property
    def name(self) -> str:
        return self._name

    @property
    def url(self) -> str:
        return self._url

    def __str__(self):
        return self.name

    async def fetch(self) -> bytes:
        return await self._md.session.get_bytes(self._url)

    # TODO dump submission
    """
    async def dump(self, file: Union[str, bytes, PathLike, AsyncBufferedIOBase]):
        content = await self.fetch()

        if isinstance(file, AsyncBufferedIOBase):
            await file.write(content)
        else:
            async with aiofiles.open(file, 'wb') as fp:
                await fp.write(content)
    """



import asyncio
from asyncio import Lock
from time import time


class RequestLimiter:
    def __init__(self, wait_time: float):
        self._lock: Lock = Lock()
        self._wait_time: float = wait_time
        self._last_request: float = 0.0

    async def __aenter__(self):
        await self._lock.__aenter__()

        diff = self._last_request + self._wait_time - time()
        if diff > 0:
            await asyncio.sleep(diff)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._last_request = time()
        await self._lock.__aexit__(exc_type, exc_val, exc_tb)



from typing import Optional, Tuple

from ..Stage import Stage


class Spawner(Stage):
    def __init__(self, server_name: str = None):
        super().__init__()
        self.server_name: Optional[str] = server_name

    def _get(self, successors: Tuple[Stage, ...], **kwargs) -> str:
        params = {
            **{
                k: self.encode(v)
                for k, v in kwargs.items()
            }
        }

        if self.server_name is not None:
            params['server_name'] = self.encode(self.server_name),

            if len(successors) > 0:
                params['next'] = self.encode(
                    f'/hub/user-redirect/{self.server_name}/{successors[0].url(successors[1:])}')

        else:
            if len(successors) > 0:
                params['next'] = self.encode(f'/hub/user-redirect/{successors[0].url(successors[1:])}')

        query = '&'.join(f'{k}={v}' for k, v in params.items())
        return f'/hub/spawn?{query}'

    def url(self, successors: Tuple[Stage, ...]) -> str:
        return self._get(successors)



from typing import Dict, Tuple

from ..Stage import Stage


class Application(Stage):
    def __init__(self, path: str, **kwargs):
        super().__init__()

        self.path: str = path
        self.kwargs: Dict = kwargs

    def url(self, successors: Tuple[Stage, ...]) -> str:
        path = f'{self.path}'

        if self.kwargs:
            params = '&'.join(f'{key}={self.encode(value)}' for key, value in self.kwargs.items())
            path += f'?{params}'

        return path



from .. import Moodle
from .Submission import Submission
from .SubmissionFile import SubmissionFile


class Assignment:
    def __init__(self, md: Moodle, id: int):
        self._md: Moodle = md
        self._id: int = id

    @property
    def id(self) -> int:
        return self._id

    async def submissions(self) -> list[Submission]:
        # We have to set the filter to `requiregrading` to only receive
        # new submissions that have not been graded yet.
        users_per_page = 20

        # request page and extract required form parameters
        page = await self._md.session.get_html(f'/mod/assign/view.php?id={self._id}&action=grading')
        form = page.find('form', attrs={'class': 'gradingoptionsform'})

        data = {
            'action': 'saveoptions',
            'filter': 'requiregrading',
            'perpage': users_per_page
        }
        for key in 'id', 'userid', 'contextid', 'sesskey', \
                'showonlyactiveenrol', 'downloadasfolders', \
                '_qf__mod_assign_grading_options_form', 'mform_isexpanded_id_general':
            data[key] = form.find('input', attrs={'name': key})['value']

        assert data['id'] == str(self._id)

        # send post request and extract table
        page = await self._md.session.post_html('/mod/assign/view.php', data)
        table = page.find('div', attrs={'class': 'gradingtable'}).find('table')

        if table is None:
            return []

        # find users
        assert len(table.select('tbody tr')) == users_per_page

        submissions: list[Submission] = []
        for row in table.select('tbody tr:not(.emptyrow)'):
            # extract user id
            for class_name in row['class']:
                if class_name.startswith('user'):
                    user_id = int(class_name[4:])
                    break
            else:
                # TODO raise MoodleParseError
                raise AssertionError

            # extract grade attempt
            grade_attempt = row.find('input', attrs={'name': f'gradeattempt_{user_id}'})['value']

            # extract attached files
            files: list[SubmissionFile] = []
            for submission in row.find_all('div', attrs={'class': 'fileuploadsubmission'}):
                anchor = submission.find('a')
                assert anchor['href'].endswith('?forcedownload=1')

                files.append(SubmissionFile(self._md, anchor.text.strip(), anchor['href']))

            # add submission to result list
            submissions.append(Submission(self._md, self, user_id, files, grade_attempt, data['sesskey']))

        return submissions



from .MoodleError import MoodleError


class MoodleUpdateError(MoodleError):
    def __init__(self, attribute: str, should_value: str | bool, is_value: str | bool):
        super().__init__(f'{attribute} is {is_value} but should be {should_value}')

        self.attribute: str = attribute
        self.should_value: str | bool = should_value
        self.is_value: str | bool = is_value



from aiohttp import ClientSession, ClientResponse
from bs4 import BeautifulSoup
from yarl import URL

from .RequestLimiter import RequestLimiter


class LimitedSession:
    def __init__(self, host: str, wait_time: float):
        # initialize session
        connector = aiohttp.TCPConnector(
            limit=5,
            limit_per_host=1
        )

        self._session = ClientSession(connector=connector)

        # store other properties
        self.host: str = host
        self._limiter: RequestLimiter = RequestLimiter(wait_time)

    async def __aenter__(self):
        await self._session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._session.__aexit__(exc_type, exc_val, exc_tb)

    def _url(self, path: str):
        if path.startswith('https://') or path.startswith('http://'):
            return path
        elif path.startswith('/'):
            return f'{self.host}{path}'
        else:
            return f'{self.host}/{path}'

    @staticmethod
    async def _parse(response: ClientResponse) -> BeautifulSoup:
        return BeautifulSoup((await response.read()).decode('utf-8'), 'html.parser')

    async def get_html(self, path: str) -> BeautifulSoup:
        async with self._limiter:
            async with self._session.get(self._url(path)) as response:
                return await self._parse(response)

    async def get_html_redirects(self, path: str) -> tuple[BeautifulSoup, list[tuple[int, URL]]]:
        async with self._limiter:
            async with self._session.get(self._url(path)) as response:
                return (
                    await self._parse(response),
                    [(h.status, h.url) for h in response.history] + [(response.status, response.url)]
                )

    async def get_json(self, path: str) -> dict | list:
        async with self._limiter:
            async with self._session.get(self._url(path)) as response:
                return await response.json()

    async def get_bytes(self, path: str) -> bytes:
        async with self._limiter:
            async with self._session.get(self._url(path)) as response:
                return await response.read()

    async def post_html(self, path: str, data: dict[str, any]) -> BeautifulSoup:
        for key in data:
            if not isinstance(data[key], str):
                data[key] = str(data[key])

        async with self._limiter:
            async with self._session.post(self._url(path), data=data) as response:
                return await self._parse(response)

    async def post_json(self, path: str, data: dict[str, any], no_check_mimetype: bool = False) -> dict | list:
        async with self._limiter:
            async with self._session.post(self._url(path), data=data) as response:
                if not no_check_mimetype:
                    return await response.json()
                else:
                    return await response.json(content_type=None)



from setuptools import setup, find_packages

VERSION = '0.1.5'

setup(
    name='tui-moodle-api',
    version=VERSION,
    author='Eric Tröbs',
    author_email='eric.troebs@tu-ilmenau.de',
    description='scraper for moodle',
    long_description='',
    long_description_content_type='text/markdown',
    url='https://dbgit.prakinf.tu-ilmenau.de/ertr8623/moodle-api',
    project_urls={
        'Bug Tracker': 'https://dbgit.prakinf.tu-ilmenau.de/ertr8623/moodle-api/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.10',
    install_requires=[
        'aiohttp~=3.11.16',
        'beautifulsoup4~=4.13.3',
        'pyyaml~=6.0.2'
    ],
)



from bs4 import BeautifulSoup

from .. import Moodle
from ..error import MoodleParseError, MoodleUpdateError


class Activity:
    def __init__(self, md: Moodle, id: int, page: BeautifulSoup = None, **kwargs: bool | str):
        self._md: Moodle = md
        self._id: int = id
        self._page: BeautifulSoup | None = page
        self._kwargs: dict[str, bool | str] = kwargs
        self._update: dict[str, bool | str] = {}

    @property
    def id(self) -> int:
        return self._id

    @property
    def session_key(self) -> str:
        return self._kwargs['sesskey']

    async def fetch(self, inplace: bool = False) -> 'Activity':
        # fetch edit page
        self._page = await self._md.session.get_html(f'/course/modedit.php?update={self._id}')

        # find form and extract parameters
        form = self._page.select_one('form[action="modedit.php"]')

        self._kwargs = {
            **{
                text_input.attrs['name']: text_input.attrs['value']
                for text_input in form.select('input[type="text"]')
            },
            **{
                checkbox_input.attrs['name']: 'checked' in checkbox_input.attrs
                for checkbox_input in form.select('input[type="checkbox"]')
            },
            **{
                textarea.attrs['name']: textarea.text
                for textarea in form.select('textarea')
            },
            **{
                select.attrs['name']: selected_option.attrs['value']
                for select, selected_option in [
                    (select, select.select_one('option[selected]'))
                    for select in form.select('select')
                ]
                if selected_option is not None
            },
            **{
                hidden_input.attrs['name']: hidden_input.attrs['value']
                for hidden_input in form.select('input[type="hidden"]')
            },
        }

        # find type and create specific activity object
        if inplace:
            return self

        nav_links = self._page.select('a.nav-link[role="menuitem"]')

        for nl in nav_links:
            if nl.attrs['href'].endswith(f'view.php?id={self._id}'):
                match nl.text.strip().lower():
                    case 'datei':
                        from .ResourceActivity import ResourceActivity
                        return ResourceActivity(self._md, self._id, self._page, **self._kwargs)
                    case 'externes tool':
                        from .LTIActivity import LTIActivity
                        return LTIActivity(self._md, self._id, self._page, **self._kwargs)

        # raise error if type is unknown
        raise MoodleParseError(f'could not determine activity type for id={self._id}')

    def __contains__(self, key) -> bool:
        return key in self._kwargs

    def __getitem__(self, key: str) -> bool | str:
        return self._kwargs[key]

    async def set(self, **kwargs: str | bool) -> 'Activity':
        # If activity has not been fetched, we need to do it now.
        if self._page is None:
            await self.fetch(inplace=True)

        # send all properties
        await self._md.session.post_html(f'/course/modedit.php', data={
            **{
                k: v if not isinstance(v, bool) else (1 if v else 0)
                for k, v in {
                    **self._kwargs,
                    **self._update,
                    **kwargs
                }.items()
            },
            'submitbutton': 'Speichern und anzeigen'
        })

        self._update = {}

        # fetch again and check the updated properties
        await self.fetch(inplace=True)

        for k, v in kwargs.items():
            if v != self._kwargs[k]:
                raise MoodleUpdateError(k, v, self._kwargs[k])

        return self

    def store(self, **kwargs: str | bool) -> 'Activity':
        for k, v in kwargs.items():
            self._update[k] = v

        return self

    async def _set(self, _autosave: bool, **kwargs: str | bool):
        if _autosave:
            return await self.set(**kwargs)
        else:
            return self.store(**kwargs)

    async def set_name(self, name: str, save: bool = False) -> 'Activity':
        return await self._set(save, name=name)

    async def set_description(self, description: str, description_format: int = 2, save: bool = False) -> 'Activity':
        return await self._set(save, **{
            'introeditor[text]': description,
            'introeditor[format]': str(description_format),
        })



from .Application import Application


class VSCode(Application):
    def __init__(self, folder: str = None):
        if folder is not None:
            super().__init__('vscode', folder=folder)
        else:
            super().__init__('vscode')



from .Stage import Stage
from .action import *
from .application import *
from .spawner import *
from .targets import *


def generate(hub: str, image: str, git_repo: str, git_branch: str, git_target: str, start: str) -> str:
    hub = JupyterHub(hub)
    image = hub + DockerSpawner(image)
    git = image + NBGitPuller(git_repo, git_branch, git_target)
    lab = git + JupyterLab(f'{git_target}/{start}')

    return str(lab)
